/*
 * Read a range or all PCR registers of a TPM
 *
 * Copyright (C) 2011 Andreas Steffen
 *
 * HSR Hochschule fuer Technik Rapperswil
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <trousers/tss.h>

#define CKERR	if (result != TSS_SUCCESS) goto error

BYTE hex_to_bin(char *p)
{
	BYTE c = *p;
	
	if (c >= '0' && c <= '9')
	{
		return c - '0';
	}
	else if (c >= 'A' && c <= 'F')
	{
		return c - 'A' + 10;
	}
	else if (c >= 'a' && c <= 'f')
	{
		return c - 'a' + 10;
	}
	else
	{
		return 0xff;
	}
}

int main (int ac, char **av)
{
	TSS_HCONTEXT	hContext;
	TSS_HTPM	hTPM;
	UINT32		pcrMaxIndex, pcrIndex, pcrValueLen, pcrDataLen;
	UINT32		respDataLen, subCap;
	BYTE		*pcrData, *pcrValue, *respData;
	char		*pcrHexData;
	int			pcrHexDataLen, result, i;

	if (ac != 3 || (strcmp(av[1], "-h") == 0 || strcmp(av[1], "--help") == 0))
	{
		fprintf(stderr, "Usage: %s pcrIndex pcrHexData\n", av[0]);
		exit (0);
	}

	/* Initialize TPM context */
	result = Tspi_Context_Create(&hContext); CKERR;
	result = Tspi_Context_Connect(hContext, NULL); CKERR;
	result = Tspi_Context_GetTpmObject (hContext, &hTPM); CKERR;

	/* Get the numbe of PCR registers */
	subCap = TSS_TPMCAP_PROP_PCR;
	result = Tspi_TPM_GetCapability(hTPM, TSS_TPMCAP_PROPERTY,
									4, (BYTE*)&subCap,
									&respDataLen, &respData); CKERR;
	if (respDataLen != 4)
	{
		fprintf(stderr, "Failure: "
						"number of PCR registers is not of type UINT32\n");
		exit(1);
	}
	pcrMaxIndex = *((UINT32*)respData) - 1;
	pcrIndex = atoi(av[1]);
	pcrHexData = av[2];
	pcrHexDataLen = strlen(pcrHexData);

	/* Check if the PCR to be extended exists */
	if (pcrIndex > pcrMaxIndex)
	{
		fprintf(stderr, "Error: pcsIndex > maxIndex of %u\n",
				pcrMaxIndex);
		exit(1);
	}

	/* Check if the data to be extended has an even number of HEX symbols */
	if (pcrHexDataLen % 2)
	{
		fprintf(stderr, "Error: pcsHexData has an odd number of %d HEX symbols\n",
			    pcrHexDataLen);
		exit(1);
	}

	pcrDataLen = pcrHexDataLen / 2;
	pcrData = malloc(pcrDataLen);

	/* Convert the PCR data from HEX to binary format */
	for (i = 0; i < pcrDataLen; i++)
	{
		BYTE high_nibble, low_nibble;

		high_nibble = hex_to_bin(pcrHexData + 2*i);
		low_nibble  = hex_to_bin(pcrHexData + 2*i + 1);

		if (high_nibble == 0xff || low_nibble == 0xff)
		{
			fprintf(stderr, "Error: non-HEX symbol in pcrData\n");
			exit(1);
		}			
		pcrData[i] = 16*high_nibble + low_nibble;
	}

	/* Extend the PCR with the data and output extended PCR value */
	result = Tspi_TPM_PcrExtend(hTPM, pcrIndex, pcrDataLen, pcrData,
								NULL, &pcrValueLen, &pcrValue); CKERR;

	printf("pcr%02u = ", pcrIndex);
	for (i = 0;  i < pcrValueLen; i++)
	{
		printf("%02x", pcrValue[i]);
	}
	printf("\n");
	result = Tspi_Context_FreeMemory(hContext, pcrValue); CKERR;

	exit(0);

error:
	fprintf (stderr, "Failure, error code: 0x%x\n", result);
	return 1;
}
